"use strict";Object.defineProperty(exports, "__esModule", {value: true}); function _nullishCoalesce(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } }var _biggystring = require('biggystring');









var _cleaners = require('cleaners');










var _types = require('edge-core-js/types');
var _ethers = require('ethers');






var _swapHelpers = require('../../util/swapHelpers');







var _utils = require('../../util/utils');

var _abiMap = require('./abi/abiMap');
var _defiUtils = require('./defiUtils');











var _thorchain = require('./thorchain');

const pluginId = 'thorchainda'
const swapInfo = {
  pluginId,
  isDex: true,
  displayName: 'Thorchain DEX Aggregator',
  supportEmail: 'support@edge.app'
}

// This needs to be a type so adding the '& {}' prevents auto correction to an interface











const asCalldata = _cleaners.asObject.call(void 0, {
  tcMemo: _cleaners.asOptional.call(void 0, _cleaners.asString),
  memo: _cleaners.asOptional.call(void 0, _cleaners.asString)
})

const asThorSwapRoute = _cleaners.asObject.call(void 0, {
  contract: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  contractMethod: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  contractInfo: _cleaners.asOptional.call(void 0, _cleaners.asString),
  complete: _cleaners.asBoolean,
  path: _cleaners.asString,
  providers: _cleaners.asArray.call(void 0, _cleaners.asString),
  calldata: _cleaners.asUnknown,
  expectedOutput: _cleaners.asString,
  expectedOutputMaxSlippage: _cleaners.asString,
  expectedOutputUSD: _cleaners.asString,
  expectedOutputMaxSlippageUSD: _cleaners.asString,
  deadline: _cleaners.asOptional.call(void 0, _cleaners.asString)
})

const asThorSwapQuoteResponse = _cleaners.asObject.call(void 0, {
  routes: _cleaners.asArray.call(void 0, asThorSwapRoute)
})

const DA_VOLATILITY_SPREAD_DEFAULT = 0.03
const THORSWAP_DEFAULT_SERVERS = ['https://api.thorswap.net/aggregator']



let exchangeInfo
let exchangeInfoLastUpdate = 0

const tokenProxyMap = {
  ethereum: '0xf892fef9da200d9e84c9b0647ecff0f34633abe8',
  avalanche: '0x69ba883af416ff5501d54d5e27a1f497fbd97156'
}

 function makeThorchainDaPlugin(
  opts
) {
  const { io, log } = opts
  const { fetchCors = io.fetch } = io
  const {
    appId,
    affiliateFeeBasis,
    ninerealmsClientId,
    thorname,
    thorswapApiKey
  } = _thorchain.asInitOptions.call(void 0, opts.initOptions)

  const headers = {
    'Content-Type': 'application/json',
    'x-client-id': ninerealmsClientId
  }

  const thorswapHeaders = {
    'Content-Type': 'application/json',
    referer: thorswapApiKey
  }

  const fetchSwapQuoteInner = async (
    request
  ) => {
    const {
      fromCurrencyCode,
      toCurrencyCode,
      nativeAmount,
      fromWallet,
      fromTokenId,
      toWallet,
      toTokenId,
      quoteFor
    } = request
    // Do not support transfer between same assets
    if (
      fromWallet.currencyInfo.pluginId === toWallet.currencyInfo.pluginId &&
      request.fromCurrencyCode === request.toCurrencyCode
    ) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }
    const reverseQuote = quoteFor === 'to'
    const isEstimate = false

    let daVolatilitySpread = DA_VOLATILITY_SPREAD_DEFAULT
    let thornodeServers = _thorchain.THORNODE_SERVERS_DEFAULT
    let thorswapServers = THORSWAP_DEFAULT_SERVERS

    _swapHelpers.checkInvalidCodes.call(void 0, _thorchain.INVALID_CURRENCY_CODES, request, swapInfo)

    // Grab addresses:
    const fromAddress = await _utils.getAddress.call(void 0, fromWallet)
    const toAddress = await _utils.getAddress.call(void 0, toWallet)

    const fromMainnetCode =
      _thorchain.MAINNET_CODE_TRANSCRIPTION[fromWallet.currencyInfo.pluginId]
    const toMainnetCode =
      _thorchain.MAINNET_CODE_TRANSCRIPTION[toWallet.currencyInfo.pluginId]

    if (fromMainnetCode == null || toMainnetCode == null) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    const now = Date.now()
    if (
      now - exchangeInfoLastUpdate > _thorchain.EXCHANGE_INFO_UPDATE_FREQ_MS ||
      exchangeInfo == null
    ) {
      try {
        const exchangeInfoResponse = await _utils.promiseWithTimeout.call(void 0, 
          _utils.fetchInfo.call(void 0, fetchCors, `v1/exchangeInfo/${appId}`)
        )

        if (exchangeInfoResponse.ok === true) {
          exchangeInfo = _thorchain.asExchangeInfo.call(void 0, await exchangeInfoResponse.json())
          exchangeInfoLastUpdate = now
        } else {
          // Error is ok. We just use defaults
          log.warn('Error getting info server exchangeInfo. Using defaults...')
        }
      } catch (e) {
        log.warn(
          'Error getting info server exchangeInfo. Using defaults...',
          e.message
        )
      }
    }

    if (exchangeInfo != null) {
      const { thorchain } = exchangeInfo.swap.plugins
      daVolatilitySpread = thorchain.daVolatilitySpread
      thorswapServers = _nullishCoalesce(thorchain.thorSwapServers, () => ( THORSWAP_DEFAULT_SERVERS))
      thornodeServers = _nullishCoalesce(thorchain.thornodeServers, () => ( thornodeServers))
    }

    const volatilitySpreadFinal = daVolatilitySpread // Might add a likeKind spread later

    //
    // Get Quote
    //
    if (reverseQuote) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    const sellAmount = await fromWallet.nativeToDenomination(
      nativeAmount,
      fromCurrencyCode
    )

    const quoteParams = {
      sellAsset:
        `${fromMainnetCode}.${fromCurrencyCode}` +
        (fromTokenId != null ? `-0x${fromTokenId}` : ''),
      buyAsset:
        `${toMainnetCode}.${toCurrencyCode}` +
        (toTokenId != null ? `-0x${toTokenId}` : ''),
      sellAmount,
      slippage: (volatilitySpreadFinal * 100).toString(),
      recipientAddress: toAddress,
      senderAddress: fromAddress,
      affiliateAddress: thorname,
      affiliateBasisPoints: affiliateFeeBasis
    }
    const sourceTokenContractAddress =
      fromTokenId != null ? `0x${fromTokenId}` : undefined
    const queryParams = _utils.makeQueryParams.call(void 0, quoteParams)
    const uri = `tokens/quote?${queryParams}`

    log.warn(uri)

    const [iaResponse, thorSwapResponse] = await Promise.all([
      _utils.fetchWaterfall.call(void 0, 
        fetchCors,
        thornodeServers,
        'thorchain/inbound_addresses',
        {
          headers
        }
      ),
      _utils.fetchWaterfall.call(void 0, fetchCors, thorswapServers, uri, {
        headers: thorswapHeaders
      })
    ])

    if (!iaResponse.ok) {
      const responseText = await iaResponse.text()
      throw new Error(
        `Thorchain could not fetch inbound_addresses: ${JSON.stringify(
          responseText,
          null,
          2
        )}`
      )
    }
    if (!thorSwapResponse.ok) {
      const responseText = await thorSwapResponse.text()
      throw new Error(
        `Thorchain could not get thorswap quote: ${JSON.stringify(
          responseText,
          null,
          2
        )}`
      )
    }

    const iaJson = await iaResponse.json()
    const inboundAddresses = _thorchain.asInboundAddresses.call(void 0, iaJson)

    const thorSwapJson = await thorSwapResponse.json()
    const thorSwapQuote = asThorSwapQuoteResponse(thorSwapJson)

    // Check for supported chain and asset
    const inAddressObject = inboundAddresses.find(
      addrObj => !addrObj.halted && addrObj.chain === fromMainnetCode
    )
    if (inAddressObject == null) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }
    const { router, address: thorAddress } = inAddressObject
    const { routes } = thorSwapQuote
    const [thorSwap] = routes

    if (thorSwap == null) throw new (0, _types.SwapCurrencyError)(swapInfo, request)

    const {
      providers,
      path,
      contractMethod,
      expectedOutputMaxSlippage
    } = thorSwap

    const calldata = asCalldata(thorSwap.calldata)

    if (providers.length <= 1) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    const tcDirect = providers[0] === 'THORCHAIN'

    const toNativeAmount = _biggystring.toFixed.call(void 0, 
      await toWallet.denominationToNative(
        expectedOutputMaxSlippage,
        toCurrencyCode
      ),
      0,
      0
    )

    // let customNetworkFee
    // let customNetworkFeeKey

    // const customFeeTemplate = (fromWallet.currencyInfo.customFeeTemplate ??
    //   [])[0]
    // const fromCurrencyInfo = fromWallet.currencyInfo
    // if (customFeeTemplate?.type === 'nativeAmount') {
    //   customNetworkFee = inAssetGasRate
    //   customNetworkFeeKey = customFeeTemplate.key
    // } else if (fromCurrencyInfo.defaultSettings?.customFeeSettings != null) {
    //   const customFeeSettings = asCustomFeeSettings(
    //     fromCurrencyInfo.defaultSettings.customFeeSettings
    //   )
    //   // Only know about the key 'gasPrice'
    //   const usesGasPrice = customFeeSettings.find(f => f === 'gasPrice')
    //   if (usesGasPrice != null) {
    //     customNetworkFee = inAssetGasRate
    //     customNetworkFeeKey = 'gasPrice'
    //   }
    // }

    // if (customNetworkFee == null || customNetworkFeeKey == null) {
    //   throw new SwapCurrencyError(swapInfo, request)
    // }

    let memo = _nullishCoalesce(_nullishCoalesce(calldata.tcMemo, () => ( calldata.memo)), () => ( ''))

    log.warn(memo)

    const contractAddress = tcDirect ? router : thorSwap.contract
    const calldataAny = thorSwap.calldata
    let ethNativeAmount = nativeAmount
    let publicAddress = thorAddress
    let approvalData

    if (_thorchain.EVM_CURRENCY_CODES[fromMainnetCode]) {
      if (fromMainnetCode !== fromCurrencyCode) {
        if (contractAddress == null)
          throw new Error(`Missing router address for ${fromMainnetCode}`)
        if (sourceTokenContractAddress == null)
          throw new Error(
            `Missing sourceTokenContractAddress for ${fromMainnetCode}`
          )
        // Need to use ethers.js to craft a proper tx that calls Thorchain contract, then extract the data payload
        // Token transactions send no ETH (or other EVM mainnet coin)
        if (tcDirect) {
          memo = await _defiUtils.getEvmTokenData.call(void 0, {
            assetAddress: sourceTokenContractAddress,
            amountToSwapWei: Number(nativeAmount),
            contractAddress,
            vaultAddress: thorAddress,
            memo
          })
        } else {
          if (contractMethod == null)
            throw new Error('Invalid null contractMethod')
          if (contractAddress == null)
            throw new Error('Invalid null contractAddress')

          memo = await exports.getCalldataData.call(void 0, 
            fromWallet.currencyInfo.pluginId,
            contractAddress,
            contractMethod,
            calldataAny
          )
        }

        ethNativeAmount = '0'
        publicAddress = contractAddress

        // Check if token approval is required and return necessary data field
        approvalData = await _defiUtils.getEvmApprovalData.call(void 0, {
          contractAddress: tokenProxyMap[fromWallet.currencyInfo.pluginId],
          assetAddress: sourceTokenContractAddress,
          nativeAmount
        })
      } else {
        memo = '0x' + Buffer.from(memo).toString('hex')
      }
    } else {
      // Cannot yet do tokens on non-EVM chains
      if (fromMainnetCode !== fromCurrencyCode) {
        throw new (0, _types.SwapCurrencyError)(swapInfo, request)
      }
    }

    let preTx
    if (approvalData != null) {
      if (sourceTokenContractAddress == null) {
        throw new Error('Cannot approve token w/o contract address')
      }
      const spendInfo = {
        // Token approvals only spend the parent currency
        tokenId: null,
        spendTargets: [
          {
            memo: approvalData,
            nativeAmount: '0',
            publicAddress: sourceTokenContractAddress
          }
        ],
        assetAction: {
          assetActionType: 'tokenApproval'
        },
        savedAction: {
          actionType: 'tokenApproval',
          tokenApproved: {
            pluginId: fromWallet.currencyInfo.pluginId,
            tokenId: fromTokenId,
            nativeAmount
          },
          tokenContractAddress: sourceTokenContractAddress,
          contractAddress: tokenProxyMap[fromWallet.currencyInfo.pluginId]
        }
      }
      preTx = await request.fromWallet.makeSpend(spendInfo)
    }

    const spendInfo = {
      tokenId: request.fromTokenId,
      spendTargets: [
        {
          memo,
          nativeAmount: ethNativeAmount,
          publicAddress
        }
      ],
      assetAction: {
        assetActionType: 'swap'
      },
      savedAction: {
        actionType: 'swap',
        swapInfo,
        isEstimate,
        toAsset: {
          pluginId: request.toWallet.currencyInfo.pluginId,
          tokenId: request.toTokenId,
          nativeAmount: toNativeAmount
        },
        fromAsset: {
          pluginId: request.fromWallet.currencyInfo.pluginId,
          tokenId: request.fromTokenId,
          nativeAmount: ethNativeAmount
        },
        payoutAddress: toAddress,
        payoutWalletId: toWallet.id,
        refundAddress: fromAddress
      },
      otherParams: {
        outputSort: 'targets'
      }
    }

    if (_thorchain.EVM_CURRENCY_CODES[fromMainnetCode]) {
      if (fromMainnetCode === fromCurrencyCode) {
        // For mainnet coins of EVM chains, use gasLimit override since makeSpend doesn't
        // know how to estimate an ETH spend with extra data
        const gasLimit = _thorchain.getGasLimit.call(void 0, fromMainnetCode, fromTokenId)
        if (gasLimit != null) {
          spendInfo.customNetworkFee = {
            ...spendInfo.customNetworkFee,
            gasLimit
          }
        }
      }
    }

    const providersStr = providers.join(' -> ')
    const notes = `DEX Providers: ${providersStr}\nPath: ${path}`

    return {
      request,
      spendInfo,
      swapInfo,
      fromNativeAmount: nativeAmount,
      expirationDate: new Date(Date.now() + _thorchain.EXPIRATION_MS),
      preTx,
      metadataNotes: notes
    }
  }

  const out = {
    swapInfo,

    async fetchSwapQuote(req) {
      const request = _utils.convertRequest.call(void 0, req)

      const newRequest = await _swapHelpers.getMaxSwappable.call(void 0, fetchSwapQuoteInner, request)
      const swapOrder = await fetchSwapQuoteInner(newRequest)
      return await _swapHelpers.makeSwapPluginQuote.call(void 0, swapOrder)
    }
  }
  return out
} exports.makeThorchainDaPlugin = makeThorchainDaPlugin;

const calldataOrder = {
  TC_ROUTER_GENERIC: [
    'tcRouter',
    'tcVault',
    'tcMemo',
    'token',
    'amount',
    'router',
    'data',
    'deadline'
  ],
  TC_ROUTER_UNISWAP: [
    'tcRouter',
    'tcVault',
    'tcMemo',
    'token',
    'amount',
    'amountOutMin',
    'deadline'
  ],
  TC_ROUTER_PANGOLIN: [
    'tcRouter',
    'tcVault',
    'tcMemo',
    'token',
    'amount',
    'amountOutMin',
    'deadline'
  ]
}

 const getCalldataData = async (
  currencyPluginId,
  contractAddress,
  contractMethod,
  calldata
) => {
  let abi, contractType
  try {
    const { type, data } = _abiMap.abiMap[currencyPluginId][
      contractAddress.toLowerCase()
    ]
    if (type === 'INVALID') {
      throw new Error(`Unsupported contract`)
    }
    abi = data
    contractType = type
  } catch (e) {
    throw new Error(
      `Could not find ABI for contract ${currencyPluginId}-${contractAddress}`
    )
  }

  const contractParams = calldataOrder[contractType].map(key => calldata[key])

  // initialize contract
  const contract = new _ethers.ethers.Contract(
    contractAddress,
    abi,
    _ethers.ethers.providers.getDefaultProvider()
  )

  // call the deposit method on the contract
  const tx = await contract.populateTransaction[contractMethod](
    ...contractParams
  )
  if (tx.data == null) throw new Error('No data in tx object')
  return tx.data
}; exports.getCalldataData = getCalldataData
