"use strict";Object.defineProperty(exports, "__esModule", {value: true}); function _nullishCoalesce(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } }var _cleaners = require('cleaners');













var _types = require('edge-core-js/types');









var _swapHelpers = require('../util/swapHelpers');
var _utils = require('../util/utils');


// See https://help.sideshift.ai/en/articles/4559664-which-coins-and-tokens-are-listed for list of supported currencies
const MAINNET_CODE_TRANSCRIPTION = {
  binancesmartchain: 'bsc',
  optimism: 'optimism',
  zcash: 'shielded'
}

const INVALID_CURRENCY_CODES = {
  from: {
    optimism: ['VELO'],
    polygon: ['USDC', 'USDC.e']
  },
  to: {
    polygon: ['USDC', 'USDC.e']
  }
}

const SIDESHIFT_BASE_URL = 'https://sideshift.ai/api/v2'
const pluginId = 'sideshift'
const swapInfo = {
  pluginId,
  isDex: false,
  displayName: 'SideShift.ai',
  supportEmail: 'help@sideshift.ai'
}
const ORDER_STATUS_URL = 'https://sideshift.ai/orders/'

async function checkQuoteError(
  rate,
  request,
  quoteErrorMessage
) {
  const { fromCurrencyCode, fromWallet } = request

  if (quoteErrorMessage === 'Amount too low') {
    const nativeMin = await fromWallet.denominationToNative(
      rate.min,
      fromCurrencyCode
    )
    throw new (0, _types.SwapBelowLimitError)(swapInfo, nativeMin)
  }

  if (quoteErrorMessage === 'Amount too high') {
    const nativeMax = await fromWallet.denominationToNative(
      rate.max,
      fromCurrencyCode
    )
    throw new (0, _types.SwapAboveLimitError)(swapInfo, nativeMax)
  }

  if (
    /method/i.test(quoteErrorMessage) &&
    /disabled/i.test(quoteErrorMessage)
  ) {
    throw new (0, _types.SwapCurrencyError)(swapInfo, request)
  }

  if (/country-blocked/i.test(quoteErrorMessage)) {
    throw new (0, _types.SwapPermissionError)(swapInfo, 'geoRestriction')
  }
  throw new Error(`SideShift.ai error ${quoteErrorMessage}`)
}






const createSideshiftApi = (
  baseUrl,
  fetchCors
) => {
  async function request(
    method,
    path,
    body
  ) {
    const url = `${baseUrl}${path}`

    const reply = await (method === 'GET'
      ? fetchCors(url)
      : fetchCors(url, {
          method,
          headers: {
            'Content-Type': 'application/json'
          },
          body: JSON.stringify(body)
        }))

    try {
      return await reply.json()
    } catch (e) {
      throw new Error(`SideShift.ai returned error code ${reply.status}`)
    }
  }

  return {
    get: async ( path) => await request('GET', path),
    post: async ( path, body) =>
      await request('POST', path, body)
  }
}

const fetchSwapQuoteInner = async (
  request,
  api,
  affiliateId
) => {
  const [refundAddress, settleAddress] = await Promise.all([
    _utils.getAddress.call(void 0, request.fromWallet),
    _utils.getAddress.call(void 0, request.toWallet)
  ])

  const {
    fromCurrencyCode,
    toCurrencyCode,
    fromMainnetCode,
    toMainnetCode
  } = _swapHelpers.getCodesWithTranscription.call(void 0, request, MAINNET_CODE_TRANSCRIPTION)

  const rate = asRate(
    await api.get(
      `/pair/${fromCurrencyCode}-${fromMainnetCode}/${toCurrencyCode}-${toMainnetCode}`
    )
  )

  if ('error' in rate) {
    throw new (0, _types.SwapCurrencyError)(swapInfo, request)
  }

  const permissions = asPermissions(
    await api.get('/permissions')
  )

  if (!permissions.createShift) {
    throw new (0, _types.SwapPermissionError)(swapInfo, 'geoRestriction')
  }

  const quoteAmount = await (request.quoteFor === 'from'
    ? request.fromWallet.nativeToDenomination(
        request.nativeAmount,
        request.fromCurrencyCode
      )
    : request.toWallet.nativeToDenomination(
        request.nativeAmount,
        request.toCurrencyCode
      ))

  const fixedQuoteRequest = asFixedQuoteRequest({
    depositCoin: fromCurrencyCode,
    depositNetwork: fromMainnetCode,
    settleCoin: toCurrencyCode,
    settleNetwork: toMainnetCode,
    depositAmount: request.quoteFor === 'from' ? quoteAmount : undefined,
    settleAmount: request.quoteFor === 'to' ? quoteAmount : undefined,
    affiliateId
  })

  const fixedQuote = asFixedQuote(
    await api.post('/quotes', fixedQuoteRequest)
  )

  if ('error' in fixedQuote) {
    await checkQuoteError(rate, request, fixedQuote.error.message)
    throw new Error(`SideShift.ai error ${fixedQuote.error.message}`)
  }

  const shiftRequest = asShiftRequest({
    quoteId: fixedQuote.id,
    affiliateId,
    settleAddress,
    refundAddress
  })

  const order = asOrder(
    await api.post('/shifts/fixed', shiftRequest)
  )

  if ('error' in order) {
    await checkQuoteError(rate, request, order.error.message)
    throw new Error(`SideShift.ai error ${order.error.message}`)
  }

  const amountExpectedFromNative = await request.fromWallet.denominationToNative(
    order.depositAmount,
    request.fromCurrencyCode
  )

  const amountExpectedToNative = await request.toWallet.denominationToNative(
    order.settleAmount,
    request.toCurrencyCode
  )

  const isEstimate = false

  const spendInfo = {
    tokenId: request.fromTokenId,
    spendTargets: [
      {
        nativeAmount: amountExpectedFromNative,
        publicAddress: order.depositAddress,
        uniqueIdentifier: order.depositMemo
      }
    ],
    networkFeeOption:
      request.fromCurrencyCode.toUpperCase() === 'BTC' ? 'high' : 'standard',
    assetAction: {
      assetActionType: 'swap'
    },
    savedAction: {
      actionType: 'swap',
      swapInfo,
      orderId: order.id,
      orderUri: ORDER_STATUS_URL + order.id,
      isEstimate,
      toAsset: {
        pluginId: request.toWallet.currencyInfo.pluginId,
        tokenId: request.toTokenId,
        nativeAmount: amountExpectedToNative
      },
      fromAsset: {
        pluginId: request.fromWallet.currencyInfo.pluginId,
        tokenId: request.fromTokenId,
        nativeAmount: amountExpectedFromNative
      },
      payoutAddress: settleAddress,
      payoutWalletId: request.toWallet.id,
      refundAddress
    }
  }

  return {
    request,
    spendInfo,
    swapInfo,
    fromNativeAmount: amountExpectedFromNative,
    expirationDate: _swapHelpers.ensureInFuture.call(void 0, new Date(order.expiresAt))
  }
}

const createFetchSwapQuote = (api, affiliateId) =>
  async function fetchSwapQuote(req) {
    const request = _utils.convertRequest.call(void 0, req)
    _swapHelpers.checkInvalidCodes.call(void 0, INVALID_CURRENCY_CODES, request, swapInfo)

    const newRequest = await _swapHelpers.getMaxSwappable.call(void 0, 
      fetchSwapQuoteInner,
      request,
      api,
      affiliateId
    )
    const swapOrder = await fetchSwapQuoteInner(newRequest, api, affiliateId)
    return await _swapHelpers.makeSwapPluginQuote.call(void 0, swapOrder)
  }

 function makeSideshiftPlugin(
  opts
) {
  const { io, initOptions } = opts
  const api = createSideshiftApi(SIDESHIFT_BASE_URL, _nullishCoalesce(io.fetchCors, () => ( io.fetch)))
  const fetchSwapQuote = createFetchSwapQuote(api, initOptions.affiliateId)

  return {
    swapInfo,
    fetchSwapQuote
  }
} exports.makeSideshiftPlugin = makeSideshiftPlugin;












const asError = _cleaners.asObject.call(void 0, { error: _cleaners.asObject.call(void 0, { message: _cleaners.asString }) })

const asPermissions = _cleaners.asObject.call(void 0, {
  createShift: _cleaners.asBoolean
})

const asRate = _cleaners.asEither.call(void 0, 
  _cleaners.asObject.call(void 0, {
    rate: _cleaners.asString,
    min: _cleaners.asString,
    max: _cleaners.asString,
    depositCoin: _cleaners.asString,
    depositNetwork: _cleaners.asString,
    settleCoin: _cleaners.asString,
    settleNetwork: _cleaners.asString
  }),
  asError
)

const asFixedQuoteRequest = _cleaners.asObject.call(void 0, {
  depositCoin: _cleaners.asString,
  depositNetwork: _cleaners.asString,
  settleCoin: _cleaners.asString,
  settleNetwork: _cleaners.asString,
  depositAmount: _cleaners.asOptional.call(void 0, _cleaners.asString),
  settleAmount: _cleaners.asOptional.call(void 0, _cleaners.asString),
  affiliateId: _cleaners.asString
})

const asFixedQuote = _cleaners.asEither.call(void 0, 
  _cleaners.asObject.call(void 0, {
    id: _cleaners.asString
  }),
  asError
)

const asShiftRequest = _cleaners.asObject.call(void 0, {
  quoteId: _cleaners.asString,
  affiliateId: _cleaners.asString,
  settleAddress: _cleaners.asString,
  refundAddress: _cleaners.asString
})

const asOrder = _cleaners.asEither.call(void 0, 
  _cleaners.asObject.call(void 0, {
    id: _cleaners.asString,
    expiresAt: _cleaners.asString,
    depositAddress: _cleaners.asString,
    depositMemo: _cleaners.asOptional.call(void 0, _cleaners.asString),
    settleAmount: _cleaners.asString,
    depositAmount: _cleaners.asString
  }),
  asError
)
